import re

import pytest

import requests_mock

from qaseio.pytest.plugin import QasePytestPlugin

pytest_plugins = ["pytester"]


@pytest.fixture
def mock() -> requests_mock.Mocker:
    with requests_mock.Mocker() as m:
        yield m


@pytest.fixture
def default_mocks(mock):
    def wrapper(project="PRJ", run_id=3):
        mock.get(f"/v1/project/{project}", json={"status": True, "result": {}})
        mock.get(
            f"/v1/run/{project}/{run_id}",
            json={
                "status": True,
                "result": {"cases": [i for i in range(1, 16) if i not in [3]]},
            },
        )
        mock.post(
            f"/v1/run/{project}", json={"status": True, "result": {"id": 3}},
        )

    return wrapper


@pytest.fixture
def cases_mocks(mock):
    def wrapper(regex=r".*/case/[a-zA-Z]+/\d+", **kwargs):
        mock.get(
            re.compile(regex), json={"status": True, "result": {}}, **kwargs
        )

    return wrapper


@pytest.fixture
def results_mocks(mock):
    def wrapper(regex=r".*/result/[a-zA-Z]+/\d+", **kwargs):
        mock.post(
            re.compile(regex),
            json={"status": True, "result": {"hash": "1a2b3d"}},
            **kwargs,
        )
        mock.patch(
            re.compile(regex + r"/.*"),
            json={"status": True, "result": {"hash": "1a2b3d"}},
            **kwargs,
        )

    return wrapper


@pytest.fixture
def attachment_mocks(mock):
    def wrapper(regex=r".*/attachment/[a-zA-Z]+", **kwargs):
        mock.post(
            re.compile(regex),
            json={
                "status": True,
                "result": [
                    {
                        "filename": "qaseio_pytest.py",
                        "url": "https://storage.cdn.example/filename.ext",
                        "extension": "py",
                        "hash": "d81bb2beb147c2eceddbf3e10f98e6216cc757e3",
                        "mime": "text\\/x-python",
                        "team": "c66dc393c83fe149449e5de3e64545279e1442ed",
                    }
                ],
            },
            **kwargs,
        )

    return wrapper


@pytest.fixture
def qs_plugin(default_mocks):
    def wrapper(**kwargs):
        default_mocks("PRJCODE", 123)
        kwargs["api_token"] = kwargs.get("api_token", "apitoken")
        kwargs["project"] = kwargs.get("project", "PRJCODE")
        kwargs["testrun"] = kwargs.get("testrun", 123)
        return QasePytestPlugin(**kwargs)

    return wrapper



import json
import re
from enum import Enum

import pytest

from qaseio.client.models import TestRunResultStatus
from qaseio.pytest.plugin import QasePytestPlugin, get_ids_from_pytest_nodes

PYTEST_FILE = """
from qaseio.pytest import qase

def test_no_deco():
    pass

@qase.id(1)
def test_single_id():
    pass

@qase.id(2, 3)
def test_multiple_ids():
    pass
"""

PYTEST_ALL_DECOS_FILE = """
from qaseio.pytest import qase

@qase.id(1)
def test_single_id():
    pass

@qase.id(4, 3)
def test_multiple_ids():
    pass
"""

PYTEST_COMPLEX_FILE = """
import pytest
from qaseio.pytest import qase

def test_no_deco():
    pass

@qase.id(1)
def test_single_id_fail():
    assert 0

@qase.id(2, 3)
def test_multiple_ids_fail():
    assert 0

@qase.id(4)
def test_single_id_pass():
    pass

@qase.id(5, 6)
def test_multiple_ids_pass():
    pass

@pytest.fixture(scope="session")
def setup():
    raise ValueError()

@qase.id(7)
def test_single_id_setup_fail(setup):
    pass

@qase.id(8, 9)
def test_multiple_ids_setup_fail(setup):
    pass

@pytest.fixture()
def teardown():
    yield
    raise ValueError()

@qase.id(10)
def test_single_id_teardown_fail(teardown):
    pass

@qase.id(11, 12)
def test_multiple_ids_teardown_fail(teardown):
    pass

@qase.id(13)
@pytest.mark.skip
def test_single_id_teardown_skip():
    pass

@qase.id(14, 15)
@pytest.mark.skip
def test_multiple_ids_teardown_skip():
    pass
"""


class FakeTerminalWriter:
    results = []

    def ensure_newline(self):
        pass

    def section(self, *args, **kwargs):
        pass

    def line(self, s="", **kw):
        self.results.append([s, kw])


class FakePluginManager:
    writer = None

    def set_writer(self, writer):
        self.writer = writer

    def get_plugin(self, *args, **kwargs):
        return self.writer


class FakeConf:
    pluginmanager = FakePluginManager()


@pytest.fixture()
def testfile(testdir):
    testdir.makepyfile(PYTEST_FILE)
    return testdir.getitems(PYTEST_FILE)


def test_get_ids_from_pytest_nodes(testfile):
    print(testfile)
    data, ids = get_ids_from_pytest_nodes(testfile)
    assert len(data) == 2
    assert len(ids) == 1
    assert data[testfile[1].nodeid].get("ids") == (1,)
    assert data[testfile[2].nodeid].get("ids") == (2, 3)


def test_plugin_init(qs_plugin, mock):
    qs_plugin()
    assert len(mock.request_history) == 2
    assert re.findall(r".*/project/[a-zA-Z]+", mock.request_history[0].url)
    assert re.findall(r".*/run/[a-zA-Z]+/\w+", mock.request_history[1].url)


def test_plugin_init_missing_project(mock):
    mock.get("/v1/project/123", status_code=404)
    with pytest.raises(ValueError):
        QasePytestPlugin("123", "123", None)


def test_plugin_init_missing_run_id(mock, default_mocks):
    default_mocks()
    mock.get("/v1/run/PRJ/123123", status_code=404)
    with pytest.raises(ValueError):
        QasePytestPlugin("123", "PRJ", 123123)


def test_pytest_report_header(qs_plugin, mock):
    plugin = qs_plugin()
    text = plugin.pytest_report_header(None, None)
    assert text == "qase: existing testrun #123 selected"
    plugin.testrun_id = None
    text = plugin.pytest_report_header(None, None)
    assert text == "qase: a new testrun will be created"


def test_pytest_collection_modifyitems(qs_plugin, mock, cases_mocks, testfile):
    cases_mocks()
    config = FakeConf()
    writer = FakeTerminalWriter()
    config.pluginmanager.set_writer(writer)
    plugin = qs_plugin(debug=True)
    mock.reset_mock()
    plugin.pytest_collection_modifyitems(None, config, testfile)
    assert len(mock.request_history) == 3
    assert len(writer.results) == 2
    assert len(plugin.nodes_with_ids) == 2


@pytest.mark.parametrize(
    "cases_kwargs, tests",
    [({"status_code": 404}, PYTEST_FILE), ({}, PYTEST_ALL_DECOS_FILE)],
)
def test_pytest_collection_modifyitems_missing_cases(
    qs_plugin, mock, cases_mocks, testdir, cases_kwargs, tests
):
    cases_mocks(**cases_kwargs)
    plugin = qs_plugin()
    mock.reset_mock()
    testdir.makepyfile(tests)
    items = testdir.getitems(tests)
    plugin.pytest_collection_modifyitems(None, None, items)
    assert len(mock.request_history) == 3
    assert len(plugin.nodes_with_ids) == 2


def test_start_pytest_item(qs_plugin, mock, cases_mocks, testfile):
    cases_mocks()
    plugin = qs_plugin()
    mock.reset_mock()
    plugin.pytest_collection_modifyitems(None, None, testfile)
    mock.reset_mock()
    mock.post(
        "/v1/result/PRJCODE/123",
        json={"status": True, "result": {"hash": "1a2b3d"}},
    )
    for item in testfile:
        plugin.start_pytest_item(item)
    assert len(mock.request_history) == 2
    for k, v in plugin.nodes_with_ids.items():
        assert v.get("hashes") == ["1a2b3d"]


@pytest.fixture
def start_items(qs_plugin, mock, cases_mocks, testfile):
    cases_mocks()
    plugin = qs_plugin()
    plugin.pytest_collection_modifyitems(None, None, testfile)
    mock.post(
        "/v1/result/PRJCODE/123",
        json={"status": True, "result": {"hash": "1a2b3d"}},
    )
    for item in testfile:
        plugin.start_pytest_item(item)
    mock.reset_mock()
    return plugin


def test_finish_pytest_item_no_errors(mock, start_items, testfile):
    plugin = start_items
    mock.patch(
        "/v1/result/PRJCODE/123/1a2b3d",
        json={"status": True, "result": {"hash": "1a2b3d"}},
    )
    for k, v in plugin.nodes_with_ids.items():
        v["result"] = TestRunResultStatus.PASSED
    for item in testfile:
        plugin.finish_pytest_item(item)
    assert len(mock.request_history) == 2
    for req in mock.request_history:
        data = json.loads(req.body)
        assert data.get("status") == "passed"
        assert data.get("comment") == plugin.comment


def test_finish_pytest_item_errors(mock, start_items, testfile):
    plugin = start_items
    mock.patch(
        "/v1/result/PRJCODE/123/1a2b3d",
        json={"status": True, "result": {"hash": "1a2b3d"}},
    )
    for k, v in plugin.nodes_with_ids.items():
        v["result"] = TestRunResultStatus.FAILED
        v["error"] = "error"
    for item in testfile:
        plugin.finish_pytest_item(item)
    assert len(mock.request_history) == 2
    for req in mock.request_history:
        data = json.loads(req.body)
        assert data.get("status") == "failed"
        assert data.get("comment") == plugin.comment
        assert data.get("stacktrace") == "error"
        assert data.get("time") is not None


def test_complex_run(qs_plugin, mock, default_mocks, cases_mocks, testdir):
    default_mocks()
    cases_mocks()

    def json_callback(request, context):
        data = json.loads(request.body)
        return {"status": True, "result": {"hash": str(data.get("case_id"))}}

    result = {"status": True, "result": {"hash": "1a2b3d"}}
    mock.post("/v1/result/PRJ/3", json=json_callback)
    mock.patch(re.compile(r"/v1/result/PRJ/3/[0-9]+"), json=result)
    testdir.makepyfile(PYTEST_COMPLEX_FILE)
    items = testdir.getitems(PYTEST_COMPLEX_FILE)
    testdir.runpytest(
        "-v",
        "--qase",
        "--qase-project=PRJ",
        "--qase-testrun=3",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    data, _ = get_ids_from_pytest_nodes(items)
    for nodeid, values in data.items():
        status = TestRunResultStatus.PASSED
        if ("setup" in nodeid or "teardown" in nodeid) and "fail" in nodeid:
            status = TestRunResultStatus.BLOCKED
        elif "skip" in nodeid:
            status = "skipped"
        elif "fail" in nodeid:
            status = TestRunResultStatus.FAILED
        print(nodeid, status)
        for _id in values.get("ids"):
            for req in mock.request_history:
                if req.url.endswith(f"result/PRJ/3/{_id}"):
                    data = json.loads(req.body)
                    if isinstance(status, Enum):
                        status = status.value
                    assert data.get("status") == status



from qaseio.pytest.plugin import QasePytestPluginSingleton


def get_option_ini(config, name):
    ret = config.getoption(name)  # 'default' arg won't work as expected
    if ret in (None, False):
        ret = config.getini(name)
    return ret


def pytest_addoption(parser):
    group = parser.getgroup("qase")

    def add_option_ini(option, dest, default=None, type=None, **kwargs):
        parser.addini(
            dest,
            default=default,
            type=type,
            help="default value for " + option,
        )
        group.addoption(option, dest=dest, **kwargs)

    add_option_ini(
        "--qase",
        "qs_enabled",
        default=False,
        type="bool",
        help="Use Qase TMS",
        action="store_true",
    )
    add_option_ini(
        "--qase-api-token", "qs_api_token", help="Api token for Qase TMS",
    )
    add_option_ini(
        "--qase-project", "qs_project_code", help="Project code in Qase TMS",
    )
    add_option_ini(
        "--qase-testrun", "qs_testrun_id", help="Testrun ID in Qase TMS",
    )
    add_option_ini(
        "--qase-testplan", "qs_testplan_id", help="Testplan ID in Qase TMS",
    )
    add_option_ini(
        "--qase-new-run",
        "qs_new_run",
        default=False,
        type="bool",
        help="Create new testrun, if no testrun id provided",
        action="store_true",
    )
    add_option_ini(
        "--qase-debug",
        "qs_debug",
        default=False,
        type="bool",
        help="Prints additional output of plugin",
        action="store_true",
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "qase(*ids): mark test to be associate with Qase TMS"
    )
    if get_option_ini(config, "qs_enabled"):
        QasePytestPluginSingleton.init(
            api_token=get_option_ini(config, "qs_api_token"),
            project=get_option_ini(config, "qs_project_code"),
            testrun=get_option_ini(config, "qs_testrun_id"),
            testplan=get_option_ini(config, "qs_testplan_id"),
            create_run=get_option_ini(config, "qs_new_run"),
            debug=get_option_ini(config, "qs_debug"),
        )
        config.pluginmanager.register(
            QasePytestPluginSingleton.get_instance(), name="qase-pytest",
        )



import re

TEST_FILE = """
from qaseio.pytest import qase
@qase.id(1)
def test_example():
    pass
"""


def test_run_all_parameters_cli(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(
        """
    def test_example():
        pass
    """
    )
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-testrun=3",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret == 0
    assert len(mock.request_history) == 2
    assert re.findall(r".*/project/PRJ", mock.request_history[0].url)
    assert re.findall(r".*/run/PRJ/3", mock.request_history[1].url)
    assert mock.request_history[0].headers.get("Token") == "12345"


def test_run_create_testrun_no_ids(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(
        """
    from qaseio.pytest import qase
    def test_example():
        qase.attach((b'', "text/plain", "example.txt"))
    """
    )
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-new-run",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret == 0
    assert len(mock.request_history) == 1
    assert re.findall(r".*/project/PRJ", mock.request_history[0].url)
    assert mock.request_history[0].headers.get("Token") == "12345"


def test_run_create_testrun(
    mock, default_mocks, cases_mocks, results_mocks, attachment_mocks, testdir
):
    default_mocks()
    cases_mocks()
    results_mocks()
    attachment_mocks()
    testdir.makepyfile(
        """
    from qaseio.pytest import qase
    @qase.id(1)
    def test_example():
        qase.attach((b'{}', "text/plain", "example.txt"))
    """
    )
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-new-run",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret == 0
    assert len(mock.request_history) == 7
    assert re.findall(r".*/project/PRJ", mock.request_history[0].url)
    assert mock.request_history[0].headers.get("Token") == "12345"


def test_run_create_testrun_and_id(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(TEST_FILE)
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-testrun=3",
        "--qase-new-run",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret != 0
    assert len(mock.request_history) == 2


def test_run_create_testrun_and_testplan(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(TEST_FILE)
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-testplan=3",
        "--qase-new-run",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret != 0
    assert len(mock.request_history) == 1


def test_run_testrun_and_testplan(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(TEST_FILE)
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-testplan=3",
        "--qase-testrun=4",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret != 0
    assert len(mock.request_history) == 1


def test_run_testrun_and_create_testrun_and_testplan(
    mock, default_mocks, testdir
):
    default_mocks()
    testdir.makepyfile(TEST_FILE)
    result = testdir.runpytest(
        "--qase",
        "--qase-project=PRJ",
        "--qase-testplan=3",
        "--qase-testrun=4",
        "--qase-new-run",
        "--qase-api-token=12345",
        "--qase-debug",
    )
    assert result.ret != 0
    assert len(mock.request_history) == 1


def test_run_not_enabled(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(
        """
    from qaseio.pytest import qase
    def test_example():
        qase.attach((b'', "text/plain", "example.txt"))
    """
    )
    result = testdir.runpytest()
    assert result.ret == 0
    assert not mock.called
    assert len(mock.request_history) == 0


def test_run_all_parameters_config(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(
        """
    def test_example():
        pass
    """
    )
    testdir.makefile(
        ".ini",
        pytest=(
            "[pytest]\nqs_enabled=True\nqs_project_code=PRJ\n"
            "qs_testrun_id=3\nqs_api_token=12345\nqs_debug=True"
        ),
    )
    result = testdir.runpytest()
    assert result.ret == 0
    assert len(mock.request_history) == 2
    assert re.findall(r".*/project/PRJ", mock.request_history[0].url)
    assert re.findall(r".*/run/PRJ/3", mock.request_history[1].url)
    assert mock.request_history[0].headers.get("Token") == "12345"


def test_run_override_using_cli(mock, default_mocks, testdir):
    default_mocks()
    testdir.makepyfile(
        """
    def test_example():
        pass
    """
    )
    testdir.makefile(
        ".ini",
        pytest=(
            "[pytest]\nqs_enabled=True\nqs_project_code=PRJCODE\n"
            "qs_testrun_id=1\nqs_api_token=12345\nqs_debug=True"
        ),
    )
    result = testdir.runpytest(
        "--qase-project=PRJ", "--qase-testrun=3", "--qase-api-token=45678",
    )
    assert result.ret == 0
    assert len(mock.request_history) == 2
    assert re.findall(r".*/project/PRJ", mock.request_history[0].url)
    assert re.findall(r".*/run/PRJ/3", mock.request_history[1].url)
    assert mock.request_history[0].headers.get("Token") == "45678"



import logging
import time
from datetime import datetime
from typing import Tuple, Union

import pytest

import apitist

from qaseio.client import QaseApi
from qaseio.client.models import (
    TestRunCreate,
    TestRunInclude,
    TestRunInfo,
    TestRunResultCreate,
    TestRunResultStatus,
    TestRunResultStepCreate,
    TestRunResultUpdate,
)

QASE_MARKER = "qase"
PYTEST_TO_QASE_STATUS = {
    "PASSED": TestRunResultStatus.PASSED,
    "FAILED": TestRunResultStatus.FAILED,
    "SKIPPED": TestRunResultStatus.BLOCKED,
}


def get_ids_from_pytest_nodes(items):
    """Return tuple with item and test case ids for it and tests missing ids"""
    return (
        {
            item.nodeid: {
                "ids": item.get_closest_marker(QASE_MARKER).kwargs.get("ids")
            }
            for item in items
            if item.get_closest_marker(QASE_MARKER)
        },
        [item for item in items if not item.get_closest_marker(QASE_MARKER)],
    )


class MoreThenOneCaseIdException(Exception):
    pass


class MissingStepIdentifierException(Exception):
    pass


class QasePytestPlugin:
    testrun: TestRunInfo = None

    def __init__(
        self,
        api_token,
        project,
        testrun=None,
        testplan=None,
        create_run=False,
        debug=False,
    ):
        self.client = QaseApi(api_token)
        self.project_code = project
        self.testrun_id = testrun
        self.testplan_id = testplan
        self.create_run = create_run
        self.debug = debug
        self.nodes_with_ids = {}
        self.cases_info = {}
        self.missing_ids = []
        self.existing_ids = []
        self.last_node = None
        self.project = self.client.projects.exists(self.project_code)
        self.comment = "Pytest Plugin Automation Run"
        if not self.project:
            raise ValueError("Unable to find given project code")
        self.check_testrun()
        if self.debug:
            logger = logging.getLogger(apitist.dist_name)
            logger.addHandler(logging.StreamHandler())
            logger.setLevel(logging.INFO)

    def pytest_report_header(self, config, startdir):
        """ Add extra-info in header """
        message = "qase: "
        if self.testrun_id:
            message += "existing testrun #{} selected".format(self.testrun_id)
        else:
            message += "a new testrun will be created"
        return message

    def check_case_ids(self, data):
        exist = []
        not_exist = []
        for _id in data.get("ids"):
            case = self.client.cases.exists(self.project_code, _id)
            if case:
                self.cases_info[case.id] = case
                exist.append(_id)
                continue
            not_exist.append(_id)
        return exist, not_exist

    def get_missing_in_testrun(self, data):
        missing = []
        for _id in data.get("ids"):
            if _id not in self.testrun.cases:
                missing.append(_id)
        return missing

    def check_testrun(self):
        if (self.testrun_id or self.create_run) and self.testplan_id:
            raise ValueError(
                "You should provide either use testrun or testplan"
            )
        if self.testplan_id:
            testplan = self.client.plans.exists(
                self.project_code, self.testplan_id
            )
            if not testplan:
                raise ValueError("Could not find testplan")
            self.create_testrun([case.case_id for case in testplan.cases])
        self.load_testrun()
        if not self.testrun and not self.create_run:
            raise ValueError(
                "Unable to find given testrun id, you should specify it"
            )
        if self.testrun and self.create_run:
            raise ValueError(
                "You should provide either testrun/testplan id or "
                "select to create it, not both of it"
            )

    def load_testrun(self):
        if self.testrun_id:
            self.testrun = self.client.runs.exists(
                self.project_code,
                self.testrun_id,
                include=TestRunInclude.CASES,
            )

    def create_testrun(self, cases):
        if cases:
            self.testrun_id = self.client.runs.create(
                self.project_code,
                TestRunCreate(
                    "Automated Run {}".format(str(datetime.now())),
                    cases=cases,
                ),
            ).id
            print()
            print(
                "Qase TMS: created testrun "
                "https://app.qase.io/run/{}/dashboard/{}".format(
                    self.project_code, self.testrun_id
                )
            )

    def get_step_position(self, identifier: Union[int, str], case):
        if isinstance(identifier, int):
            # We expect that this is correct position id
            return identifier
        if isinstance(identifier, str):
            # Trying to guess that identifier is hash or step name
            for pos, step in enumerate(case.steps):
                if identifier in (step.get("hash"), step.get("action")):
                    return pos + 1
        raise MissingStepIdentifierException(
            "Missing step for identifier {}".format(identifier)
        )

    def start_step(self, identifier):
        case_ids = self.nodes_with_ids[self.last_node].get("ids")
        if len(case_ids) > 1:
            # We could not match steps with more then one case
            raise MoreThenOneCaseIdException(
                "Test case decorated with several ids: {}".format(
                    ", ".join(case_ids)
                )
            )
        elif case_ids[0] in self.missing_ids:
            return None
        case = self.cases_info.get(case_ids[0])
        position = self.get_step_position(identifier, case)
        return position

    def finish_step(self, position: int, exception=None):
        if not position:
            return
        status = TestRunResultStatus.PASSED
        if exception:
            status = TestRunResultStatus.FAILED
        steps = self.nodes_with_ids[self.last_node].get("steps", {})
        steps[position] = {}
        steps[position]["status"] = status
        self.nodes_with_ids[self.last_node]["steps"] = steps

    @pytest.hookimpl(trylast=True)
    def pytest_collection_modifyitems(self, session, config, items):
        """
        Load list of items,
        check whether given ids for tests exist on server or not.

        Prints additional info at start of the run, if debug in True
        """
        self.nodes_with_ids, no_ids = get_ids_from_pytest_nodes(items)
        write_lines = []
        if no_ids:
            line = ", ".join([miss.name for miss in no_ids])
            write_lines.append(
                "This tests does not have test case ids:\n" + line
            )

        for nodeid, data in self.nodes_with_ids.items():
            exist_ids, missing_ids = self.check_case_ids(data)
            if missing_ids:
                write_lines.append(
                    "For test {} could not find test cases in TMS: {}".format(
                        nodeid, ", ".join([str(i) for i in missing_ids])
                    )
                )

            if not self.create_run:
                missing_in_run = self.get_missing_in_testrun(data)
                if missing_in_run:
                    write_lines.append(
                        "For test {} could not find "
                        "test cases in run: {}".format(
                            nodeid, ", ".join([str(i) for i in missing_in_run])
                        )
                    )
                self.missing_ids.extend(missing_in_run)
            self.missing_ids.extend(missing_ids)
            self.existing_ids.extend(exist_ids)

        if self.create_run and self.existing_ids:
            self.create_testrun(self.existing_ids)
            self.load_testrun()

        if write_lines and self.debug:
            writer = config.pluginmanager.get_plugin("terminalreporter")
            print()
            writer.ensure_newline()
            writer.section("Qase TMS", sep="=")
            for line in write_lines:
                writer.line(line)
            writer.section("Qase TMS setup finished", sep="=")

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item):
        self.start_pytest_item(item)
        yield
        self.finish_pytest_item(item)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        report = (yield).get_result()

        if item.nodeid in self.nodes_with_ids:

            def result(res):
                self.nodes_with_ids[item.nodeid]["result"] = res

            if report.longrepr:
                self.nodes_with_ids[item.nodeid]["error"] = report.longreprtext

            # Defining test result
            if report.when == "setup":
                result(None)

            if report.failed:
                if call.excinfo.typename != "AssertionError":
                    result(TestRunResultStatus.BLOCKED)
                else:
                    result(TestRunResultStatus.FAILED)
            elif report.skipped:
                if self.nodes_with_ids[item.nodeid]["result"] in (
                    None,
                    TestRunResultStatus.PASSED,
                ):
                    result("skipped")
            else:
                if self.nodes_with_ids[item.nodeid]["result"] is None:
                    result(TestRunResultStatus.PASSED)

    def start_pytest_item(self, item):
        if item.nodeid in self.nodes_with_ids:
            self.last_node = item.nodeid
            hashes = []
            for _id in self.nodes_with_ids[item.nodeid].get("ids", []):
                if _id not in self.missing_ids:
                    result = self.client.results.create(
                        self.project_code,
                        self.testrun_id,
                        TestRunResultCreate(
                            _id, TestRunResultStatus.IN_PROGRESS
                        ),
                    )
                    hashes.append(result.hash)
            self.nodes_with_ids[item.nodeid]["hashes"] = hashes
            self.nodes_with_ids[item.nodeid]["started_at"] = time.time()

    def finish_pytest_item(self, item):
        if item.nodeid in self.nodes_with_ids:
            results = self.nodes_with_ids[item.nodeid]
            hashes = results.get("hashes", [])
            attachments = results.get("attachments", [])
            steps = [
                TestRunResultStepCreate(position=pos, **values)
                for pos, values in results.get("steps", {}).items()
            ]
            attached = []
            if attachments:
                attached = self.client.attachments.upload(
                    self.project_code, *attachments
                )
            for hash in hashes:
                self.client.results.update(
                    self.project_code,
                    self.testrun_id,
                    hash,
                    TestRunResultUpdate(
                        status=results.get("result"),
                        comment=self.comment,
                        stacktrace=results.get("error"),
                        time=int(time.time() - results.get("started_at")),
                        attachments=[attach.hash for attach in attached],
                        steps=steps,
                    ),
                )
            self.last_node = None

    def add_attachments(
        self, *files: Union[str, Tuple[str, str], Tuple[bytes, str, str]]
    ):
        if self.last_node:
            node = self.nodes_with_ids[self.last_node]
            attachments: list = node.get("attachments", [])
            attachments.extend(files)
            node["attachments"] = attachments


class QasePytestPluginSingleton:
    __instance = None

    @staticmethod
    def init(**kwargs):
        QasePytestPluginSingleton.__instance = QasePytestPlugin(**kwargs)

    @staticmethod
    def get_instance() -> QasePytestPlugin:
        """ Static access method"""
        if QasePytestPluginSingleton.__instance is None:
            raise Exception("Init plugin first")
        return QasePytestPluginSingleton.__instance

    def __init__(self):
        """ Virtually private constructor"""
        raise Exception("Use get_instance()")
